@@ -11,21 +11,18 @@ import traceback
 import psutil
 from PySide2 import QtCore
 
+from . import namedprocess
+
 logger = logging.getLogger(__name__)
 
 
 TIMEOUT = 0.1  # 100 milliseconds
 
 # globals
-local_cleanup = []  # list of (func, [args], {kwargs})
-got_local_exit = False
+device_cleanup = {}  # key: device, value: list of (func, [args], {kwargs})
 device_lock = None
-proxy_ids = {}
-device_identifiers = {}  # key deivce, value: identifier
-
-
-def noop_function(device=None):
-    return None
+subproxy_devices = {}
+device_identifiers = {}  # key deivce, value: (serial_number, proxy_string)
 
 
 def _simple_access(device, function_name, *args, **kwargs):
@@ -63,20 +60,24 @@ class OptionalDeviceStringFilter:
 
     def filter(self, record):
         try:
-            record.optdevice = self.fmt % record.device
+            record.optdevice = self.fmt % record.proxy_string
         except AttributeError:
             record.optdevice = self.fallback
         return True
 
 
 class DeviceLoggerAdapter(logging.LoggerAdapter):
-    def __init__(self, logger, device_id):
-        super().__init__(logger, {'device': device_id})
+    def __init__(self, logger, serial_number, proxy_string=None):
+        if not proxy_string:
+            proxy_string = serial_number
+        super().__init__(logger, {'serial_number': serial_number,
+                                  "proxy_string": proxy_string})
 
 
 def get_device_logger(logger, device):
-    identifier = device_identifiers.get(device, "unknown")
-    return DeviceLoggerAdapter(logger, identifier)
+    serial_number, proxy_string = device_identifiers.get(
+        device, ("unknown", "unknown"))
+    return DeviceLoggerAdapter(logger, serial_number, proxy_string)
 
 
 def setup_remote_logging(log_queue):
@@ -89,24 +90,47 @@ def setup_remote_logging(log_queue):
     pyusb_logger.propagate = False
 
 
-def register_local_cleanup(cleanup_func, *args, **kwargs):
-    global local_cleanup
-    local_cleanup.append((cleanup_func, args, kwargs))
+def register_device_cleanup(device, cleanup_func, *args, **kwargs):
+    cleanup_list = device_cleanup.setdefault(device, [])
+    cleanup_list.append((cleanup_func, args, kwargs))
 
 
-def unregister_local_cleanup(cleanup_func, *args, **kwargs):
-    global local_cleanup
-    local_cleanup.remove((cleanup_func, args, kwargs))
+def unregister_device_cleanup(device, cleanup_func, *args, **kwargs):
+    cleanup_list = device_cleanup.setdefault(device, [])
+    try:
+        cleanup_list.remove((cleanup_func, args, kwargs))
+    except ValueError:
+        pass  # was already removed
 
 
-def local_exit():
-    global got_local_exit
-    got_local_exit = True
+def do_device_cleanup(device):
+    cleanup_list = device_cleanup.pop(device, [])
+    for cleanup in cleanup_list:
+        try:
+            func, args, kwargs = cleanup
+            func(device, *args, **kwargs)
+        except Exception:
+            device_logger = get_device_logger(logger, device)
+            device_logger.exception("Exception during cleanup")
 
 
-def proxy_process(log_queue, incoming, outgoing, identifier,
+def do_final_cleanup(device):
+    # do subproxy cleanup first
+    subproxy_devices = set(device_cleanup.keys())
+    subproxy_devices.discard(device)
+    subproxy_devices.discard(None)
+    for subproxy_device in subproxy_devices:
+        do_device_cleanup(subproxy_device)
+
+    # do device cleanup
+    do_device_cleanup(device)
+
+    # do final cleanup
+    do_device_cleanup(None)
+
+
+def proxy_process(log_queue, incoming, outgoing, serial_number, proxy_string,
                   find_func, ffargs, ffkwargs):
-    global local_cleanup
     global device_lock
 
     # fix a bug with stderr and stdout being None
@@ -119,7 +143,7 @@ def proxy_process(log_queue, incoming, outgoing, identifier,
     device_lock = threading.Lock()
 
     # create a logger using the device specific identifier
-    device_logger = DeviceLoggerAdapter(logger, identifier)
+    device_logger = DeviceLoggerAdapter(logger, serial_number, proxy_string)
 
     # ctrl+c handling: want to let the main process send the exit command
     if sys.platform == "win32":
@@ -144,18 +168,16 @@ def proxy_process(log_queue, incoming, outgoing, identifier,
             outgoing.put(e)
         device_logger.debug("Proxy running")
         if device:
-            proxy_ids[None] = device
-            device_identifiers[device] = identifier
+            subproxy_devices[None] = device
+            device_identifiers[device] = (serial_number, proxy_string)
             while me.parent() is not None:
                 try:
-                    if got_local_exit:
-                        break
                     job = incoming.get(True, TIMEOUT)
                     if job is None:  # check for sentinel value
                         break
-                    job_id, proxy_id, function, args, kwargs = job
+                    job_id, subproxy_id, function, args, kwargs = job
                     with device_lock:
-                        proxy_device = proxy_ids[proxy_id]
+                        proxy_device = subproxy_devices[subproxy_id]
                         proxy_logger = get_device_logger(logger, proxy_device)
                         proxy_logger.debug("got: {}".format(job))
                         try:
@@ -175,14 +197,8 @@ def proxy_process(log_queue, incoming, outgoing, identifier,
         outgoing.close()
 
         if device:
-            # make a copy of the list in case it gets modified in the loop
-            for cleanup in local_cleanup[:]:
-                try:
-                    with device_lock:
-                        func, args, kwargs = cleanup
-                        func(*args, **kwargs)
-                except Exception as e:
-                    device_logger.exception("Exception during cleanup")
+            with device_lock:
+                do_final_cleanup(device)
     except Exception:
         device_logger.exception("Uncaught Exception in Remote Process")
         raise
@@ -201,22 +217,22 @@ class DeviceProxy(QtCore.QObject):
     disconnected = QtCore.Signal()
     error = QtCore.Signal(str)
 
-    def __init__(self, log_queue, identifier, find_func, *args, **kwargs):
+    def __init__(self, process_name, log_queue, serial_number, proxy_string,
+                 find_func, *args, **kwargs):
         super().__init__()
-        self.identifier = identifier
+        self.serial_number = serial_number
+        self.proxy_string = proxy_string
         self.log_queue = log_queue
 
-        self.logger = DeviceLoggerAdapter(logger, identifier)
+        self.logger = DeviceLoggerAdapter(logger, serial_number, proxy_string)
 
         self.toProcess = multiprocessing.Queue()
         self.fromProcess = multiprocessing.Queue()
 
-        self.process = multiprocessing.Process(target=proxy_process,
-                                               args=(self.log_queue,
-                                                     self.toProcess,
-                                                     self.fromProcess,
-                                                     identifier,
-                                                     find_func, args, kwargs))
+        self.process = namedprocess.NamedProcess(
+            name=process_name, description=serial_number, target=proxy_process,
+            args=(self.log_queue, self.toProcess, self.fromProcess,
+                  serial_number, proxy_string, find_func, args, kwargs))
         self.process.daemon = True
 
         self.monitor_thread = threading.Thread(target=self.monitor)
@@ -225,7 +241,7 @@ class DeviceProxy(QtCore.QObject):
         self.jobs = {}
         self.next_job_index = 0
 
-        self.proxy_id_counter = itertools.count()
+        self.subproxy_id_counter = itertools.count()
 
     def handle_exception(self, proxy, exc):
         m = traceback.format_exception_only(type(exc), exc)
@@ -235,7 +251,9 @@ class DeviceProxy(QtCore.QObject):
         if message[-1:] == "\n":
             message = message[:-1]
 
-        proxy.logger.error(message)
+        # already logged in the proxy process
+        # proxy.logger.error(message)
+
         proxy.error.emit(message)
         proxy.close_connection()
 
@@ -312,9 +330,9 @@ class DeviceProxy(QtCore.QObject):
     def send_job(self, operation, *args, proxy=None, **kwargs):
         if proxy is None:
             proxy = self
-            proxy_id = None
+            subproxy_id = None
         else:
-            proxy_id = proxy.proxy_id
+            subproxy_id = proxy.subproxy_id
 
         job_index = self.next_job_index
         self.next_job_index = (self.next_job_index + 1) & 0xFFFFFFFF
@@ -324,13 +342,14 @@ class DeviceProxy(QtCore.QObject):
                           list(kwargs.items()))
         new_args = operation.args + args
 
-        self.toProcess.put((job_index, proxy_id, operation.function,
+        self.toProcess.put((job_index, subproxy_id, operation.function,
                             new_args, new_kwargs))
 
     def wait_for_close(self):
         if self.monitor_thread.is_alive():
             self.close_connection()
             self.monitor_thread.join()
+        QtCore.QCoreApplication.processEvents()
 
     def is_finished(self):
         if self.started:
@@ -339,28 +358,29 @@ class DeviceProxy(QtCore.QObject):
             return False
 
     def create_subproxy(self, func, *args, **kwargs):
-        proxy_id = next(self.proxy_id_counter)
-        op = DeviceOperation(create_subproxy_util, func, proxy_id,
-                             self.identifier, *args, **kwargs)
-        subproxy = DeviceSubProxy(self, proxy_id, op)
+        subproxy_id = next(self.subproxy_id_counter)
+        op = DeviceOperation(create_subproxy_util, func, subproxy_id,
+                             self.proxy_string, *args, **kwargs)
+        subproxy = DeviceSubProxy(self, subproxy_id, op)
         return subproxy
 
 
-def create_subproxy_util(device, func, proxy_id, identifier, *args, **kwargs):
+def create_subproxy_util(device, func, subproxy_id, proxy_string,
+                         *args, **kwargs):
     """ called from create_subproxy() to handle registration """
     try:
         subproxy_device = func(device, *args, **kwargs)
-        proxy_ids[proxy_id] = subproxy_device
+        subproxy_devices[subproxy_id] = subproxy_device
 
         subproxy_sn = subproxy_device.get_serial_number()
-        subproxy_identifier = "{}->{}".format(identifier, subproxy_sn)
-        device_identifiers[subproxy_device] = subproxy_identifier
+        subproxy_string = "{}->{}".format(proxy_string, subproxy_sn)
+        device_identifiers[subproxy_device] = (subproxy_sn, subproxy_string)
 
-        return (True, subproxy_identifier)
+        return (True, subproxy_sn, subproxy_string)
     except Exception:
-        message = "[{}] Unhandled exception in create_subproxy_util()"
-        logger.exception(message.format(identifier))
-        return (False, None)
+        get_device_logger(logger, device).exception(
+            "Unhandled exception in create_subproxy_util()")
+        return (False, None, None)
 
 
 class DeviceSubProxy(QtCore.QObject):
@@ -369,12 +389,12 @@ class DeviceSubProxy(QtCore.QObject):
     disconnected = QtCore.Signal()
     error = QtCore.Signal(str)
 
-    def __init__(self, proxy, proxy_id, start_op):
+    def __init__(self, proxy, subproxy_id, start_op):
         super().__init__()
 
         self.proxy = proxy
         self.logger = proxy.logger  # temporary until we get a new one
-        self.proxy_id = proxy_id
+        self.subproxy_id = subproxy_id
         self.start_op = start_op
 
         self.opened = False
@@ -383,7 +403,7 @@ class DeviceSubProxy(QtCore.QObject):
         self.proxy.disconnected.connect(self.disconnected)
         self.proxy.error.connect(self.error)
 
-        self.close_job = DeviceOperation(noop_function)
+        self.close_job = DeviceOperation(do_device_cleanup)
         self.close_job.completed.connect(self.close_completed)
 
     @QtCore.Slot()
@@ -404,9 +424,10 @@ class DeviceSubProxy(QtCore.QObject):
         self.disconnected.emit()
 
     def connected_cb(self, result):
-        success, subproxy_identifier = result
+        success, serial_number, subproxy_string = result
         if success:
-            self.logger = DeviceLoggerAdapter(logger, subproxy_identifier)
+            self.logger = DeviceLoggerAdapter(logger, serial_number,
+                                              subproxy_string)
             self.opened = True
             self.connected.emit()
         else:
@@ -434,10 +455,11 @@ class RemoteToLocalLogHandler(logging.Handler):
 
 
 class DeviceProxyManager:
-    def __init__(self):
+    def __init__(self, process_name):
+        self.process_name = process_name
         self.setup_logging()
         self.proxies = []
-        self.next_proxy_id = 0
+        self.next_proxy_number = 0
 
     def setup_logging(self):
         local_handler = RemoteToLocalLogHandler(__name__ + ".remote")
@@ -458,14 +480,16 @@ class DeviceProxyManager:
             if proxy.is_finished():
                 self.proxies.remove(proxy)
 
-    def new_proxy(self, identifier, find_func, *args, **kwargs):
+    def new_proxy(self, serial_number, find_func, *args, **kwargs):
         # clear finished proxies first
         self.clear_finished_proxies()
 
-        identifier = "{}:{}".format(self.next_proxy_id, identifier)
-        self.next_proxy_id += 1
+        proxy_number = self.next_proxy_number
+        self.next_proxy_number += 1
 
-        proxy = DeviceProxy(self.logQueue, identifier, find_func,
-                            *args, **kwargs)
+        proxy_string = "{}:{}".format(proxy_number, serial_number)
+
+        proxy = DeviceProxy(self.process_name, self.logQueue, serial_number,
+                            proxy_string, find_func, *args, **kwargs)
         self.proxies.append(proxy)
         return proxy
